
exception Unimplemented

exception NoNamedType of string
exception NoConstructor

type atype = 
  | Int (* int *)
  | Arrow of atype * atype (* arg-type, result type *)
  | Tuple of atype list (* n-tuple; ith element is ith type *)
  | Sum   of (string * atype) list (* sum with n constructors *)
  | Named of string (* named type *)
type typecontext = (string * atype) list

let rec lookup s lst exn =
  match lst with
    [] -> raise exn
  | (x,y)::tl -> if s=x then y else lookup s tl exn

(* this implementation assumes there are no recursive types; else it may
   not terminate! *)
let is_subtype (ctxt : typecontext) (t1 : atype) (t2 : atype) : bool =
  let rec r t1 t2 =
    match (t1,t2) with
      (Named s, _) -> r (lookup s ctxt (NoNamedType(s))) t2
    | (_, Named s) -> r t1 (lookup s ctxt (NoNamedType(s)))
    | (Int,Int) -> true
    | (Arrow(t1a,t1b), Arrow(t2a,t2b)) -> r t2a t1a && r t1b t2b (* contra/co *)
    | (Tuple(lst1), Tuple(lst2)) ->
	let rec loop lst1 lst2 =
	  match(lst1,lst2) with
	    (_,[]) -> true (* width *)
	  | ([],_) -> false
	  | (hd1::tl1,hd2::tl2) -> r hd1 hd2 && loop tl1 tl2 (* depth *)
	in loop lst1 lst2
    | (Sum(lst1),Sum(lst2)) ->
	let rec loop lst1 =
	  match lst1 with
	    [] -> true (* width (reversed from tuple) *)
	  | (s,t1a)::tl1 -> (* depth and permutation *)
            (try r t1a (lookup s lst2 NoConstructor)
             with NoConstructor -> false)
            && loop tl1
	in loop lst1
    | _ -> false
  in r t1 t2

(* testing -- you will probably want to add more :) *)
let c1 = [("t1",Int);("t2",Int)]
let b1 = is_subtype c1 (Named("t1")) (Named("t2"))
let b2 = is_subtype c1 (Named("t1")) (Tuple[Named("t2");Int])
let b3 = is_subtype c1 (Named("t1")) Int

let pb b = print_string(if b then "true" else "false"); print_newline()

let _ = pb b1
let _ = pb b2
let _ = pb b3
